import tqdm

#print("load " + '/'.join(__file__.split('/')[-2:]))

class ProgressBar():
    def __init__(self, iterable_obj=None, total=None, title=None, leave=False, smoothing:float=0.3, initial:int=0):
        """
        The function initializes a progress bar object with various parameters.
        
        :param iterable_obj: The `iterable_obj` parameter is used to pass an iterable object, such as a
        list or a generator, to the `__init__` method. This object will be iterated over and progress
        will be displayed using the `tqdm` library
        :param total: The `total` parameter represents the total number of iterations that will be
        performed. It is used to calculate the progress percentage and estimate the remaining time. If
        the `total` parameter is not provided or is set to `None`, the progress bar will not display the
        progress percentage or estimate the remaining time
        :param title: The `title` parameter is used to set the description of the progress bar. It is a
        string that provides a brief summary or title for the progress being tracked
        :param leave: The `leave` parameter is used to determine whether the progress bar should remain
        visible after the iteration is complete. If `leave` is set to `True`, the progress bar will
        remain visible. If `leave` is set to `False`, the progress bar will be removed after the
        iteration is complete, defaults to False (optional)
        :param smoothing: Exponential moving average smoothing factor for speed estimates (ignored in GUI mode). 
        Ranges from 0 (average speed) to 1 (current/instantaneous speed) [default: 0.3].
        :type smoothing: float
        :param initial: The `initial` parameter is used to set the initial value of the progress bar. It
        determines the starting point of the progress bar. By default, it is set to 0, defaults to 0
        :type initial: int (optional)
        """
        self.iterable = iterable_obj
        self.tqdm = tqdm.tqdm(iterable_obj, dynamic_ncols=True, total=total, leave=leave, desc=title, smoothing=smoothing, initial=initial)
        self.total = total if total != None else 0
        self.current = 0

        self.itererr = None 
        try:
            iter(self.iterable)
        except TypeError as e:
            self.itererr = e

    def Add(self, num:int=1):
        self.current = self.current + num
        self.tqdm.update(num)
    
    def Set(self, num:int):
        if num < self.current:
            raise Exception("不能小于当前进度")

        if num == self.current:
            return

        step = num - self.current
        self.current = num
        self.tqdm.update(step)
    
    def Close(self):
        self.tqdm.close()
    
    def SetTotal(self, total:int):
        self.tqdm.total = total 
        self.tqdm.refresh()
    
    def Total(self) -> int:
        return self.total 

    def Current(self) -> int:
        return self.current 
    
    def Remain(self) -> int:
        return self.total - self.current 

    def __iter__(self):
        if self.itererr != None:
            raise Exception("可迭代的参数没有传入, 需要传入, 例如Tools.ProgressBar(range(10)): " + str(self.itererr))

        for obj in self.iterable:
            # print("update")
            self.tqdm.update(1)
            yield obj 

        self.tqdm.close()
        return 
    
if __name__ == "__main__":
    import time
    for i in ProgressBar(range(10), title="test sleep"):
        time.sleep(0.3)
        #   print(i)
        
    p = ProgressBar(total=10, title="test sleep")
    for i in range(10):
        time.sleep(0.3)
        p.Add(1) 
    p.Close()


from Crypto.Cipher import AES as AAAES
from Crypto import Random
import base64

#print("load cryptoo")

class AES():
    def __init__(self, key:str, mode:str="cfb"): 
        """
        The function takes a key and a mode as arguments, and sets the block size, key, and mode of the
        cipher
        
        :param key: The key to use for the encryption
        :type key: str
        :param mode: The mode of operation for the cipher object; must be one of "ecb", "cbc", "cfb",
        "ofb", defaults to cfb
        :type mode: str (optional)
        """
        self.key = key.encode("utf-8")
        self.mode = {
            "cfb": AAAES.MODE_CFB, 
            "cbc": AAAES.MODE_CBC,
            "ecb": AAAES.MODE_ECB, 
            "ofb": AAAES.MODE_OFB,
        }[mode.lower()]
    
    def pading(self, text):
        """对加密字符的处理"""
        return text + (len(self.key) - len(text) % len(self.key)) * chr(len(self.key) - len(text) % len(self.key))

    def unpading(self, text):
        """对解密字符的处理"""
        return text[0:-ord(text[-1:])]

    def Encrypt(self, raw:str) -> str:
        # raw = self._pad(raw)
        if self.mode == AAAES.MODE_ECB:
            cipher = AAAES.new(self.key, self.mode)
            ciphertext = cipher.encrypt(bytes(self.pading(raw), encoding="utf8"))
            encrypt_string = base64.b64encode(ciphertext).decode("utf-8")
        else:
            iv = Random.new().read(AAAES.block_size)
            cipher = AAAES.new(self.key, self.mode, iv)
            ciphertext = cipher.encrypt(bytes(self.pading(raw), encoding="utf8"))
            encrypt_string = base64.b64encode(iv + ciphertext).decode("utf-8")

        return encrypt_string

    def Decrypt(self, enc:str) -> str:
        enc = base64.b64decode(enc)
        if self.mode == AAAES.MODE_ECB:
            cipher = AAAES.new(self.key, self.mode)
            plain_text = cipher.decrypt(enc)
        else:
            iv = enc[:AAAES.block_size]
            cipher = AAAES.new(self.key, self.mode, iv)
            plain_text = cipher.decrypt(enc[AAAES.block_size:])

        return self.unpading(plain_text).decode("utf-8")

if __name__ == "__main__":

    for mode in ["ecb", "cbc", "cfb", "ofb"]:
        print(mode)
        a = AES("we1d3cwged6sh6k1", mode)

        e = a.Encrypt("enc = base64.b64decode(enc)")
        print(e)

        c = a.Decrypt(e)
        print(c)
    
    print(AES("we1d3cwged6sh6k1", "ecb").Decrypt("cP9Kr1qeHXLnWBsWHVF+yFZHjy5Hq7gNl8a/1Npwu88pfDXOLnTBECVtPMQ3rgpC"))


import markdown2
import markdownify

#print("load " + __file__.split('/')[-1])

def Markdown2Html(text:str) -> str:
    extras = [
        'tables', 
        'toc', 
        'fenced-code-blocks', 
        'footnotes', 
        'task_list',
        'break-on-newline',
        'cuddled-lists',
        'strike',
        'target-blank-links'
    ]
    return markdown2.markdown(text, extras=extras)

def Html2Markdown(html:str) -> str:
    return markdownify.markdownify(html)


from __future__ import annotations

import telebot # https://github.com/eternnoir/pyTelegramBotAPI

try:
    from .Ratelimit_src import RateLimit
    from .Lock_src import Lock 
    from .DistributedLock_src import DistributedLock
    from .. import Lg
except:
    from Ratelimit_src import RateLimit
    from Lock_src import Lock
    from DistributedLock_src import DistributedLock
    import sys
    sys.path.append("..")
    import Lg

import time

#print("load " + '/'.join(__file__.split('/')[-2:]))

class TelegramBotOfficial():
    def __init__(self, token:str, ratelimit:str="20/m", lock:Lock|DistributedLock=None):
        """
        :param token: The token of your bot
        :type token: str
        :param ratelimit: The ratelimit for the bot. This is a string in the format of "x/y" where x is
        the number of messages and y is the time period. For example, "20/m" means 20 messages per
        minute, defaults to 20/m. There is no limit if set to None.
        :type ratelimit: str (optional)
        """
        self.token = token 
        self.tb = telebot.TeleBot(self.token)
        self.tags:list[str] = []
        if ratelimit != None:
            self.rl = RateLimit(ratelimit)
        else:
            self.rl = None 
        self.lock = lock
    
    def retryOnError(func): # func是被包装的函数
        def ware(self, *args, **kwargs): # self是类的实例
            errc = 0
            while True:
                try:
                    res = func(self, *args, **kwargs)
                    break
                except Exception as e:
                    Lg.Trace(str(e))
                    time.sleep(3)
                    errc += 1
                    if errc > 10:
                        raise e

            return res
    
        return ware

    def getLock(func): # func是被包装的函数
        def ware(self, *args, **kwargs): # self是类的实例
            if self.lock != None:
                self.lock.Acquire()

            res = func(self, *args, **kwargs)

            if self.lock != None:
                self.lock.Release()
            
            return res

        return ware
    
    def rateLimit(func): # func是被包装的函数
        def ware(self, *args, **kwargs): # self是类的实例
            if self.rl != None:
                self.rl.Take()

            res = func(self, *args, **kwargs)
            
            return res

        return ware

    @retryOnError
    def GetMe(self) -> telebot.types.User:
        return self.tb.get_me()
    
    def SetChatID(self, chatid:int) -> TelegramBot:
        self.chatid = chatid
        return self
    
    @retryOnError
    @getLock
    @rateLimit
    def SendFile(self, path:str):
        self.tb.send_document(self.chatid, open(path, 'rb')) 

    @retryOnError
    @getLock
    @rateLimit
    def SendImage(self, path:str):
        self.tb.send_photo(self.chatid, open(path, 'rb'))

    @retryOnError
    @getLock
    @rateLimit
    def SendVideo(self, path:str):
        self.tb.send_video(self.chatid, open(path, 'rb')) 

    @retryOnError
    @getLock
    @rateLimit
    def SendAudio(self, path:str):
        self.tb.send_audio(self.chatid, open(path, 'rb')) 

    @retryOnError
    @getLock
    @rateLimit
    def SendLocation(self, latitude:float, longitude:float):
        self.tb.send_location(self.chatid, latitude, longitude)
    
    @retryOnError
    @getLock
    @rateLimit
    def SetTags(self, *tags:str) -> TelegramBotOfficial:
        self.tags = tags
        return self 

    @retryOnError
    @getLock
    @rateLimit
    def SendMsg(self, msg:str, mode:str="text", webPreview:bool=True, *tags:str):
        """
        It sends a message to the chatid of the bot. 

        :param msg: The message to be sent
        :type msg: str
        :param mode: text, markdown, html, defaults to text
        :type mode: str (optional)
        :param webPreview: if True, the link will be shown as a link, if False, the link will be shown
        as a text, defaults to True
        :type webPreview: bool (optional)
        :param : `msg` - the message to be sent
        :type : str
        """
        if len(tags) != 0:
            tag = '\n\n' + ' '.join(['#' + t for t in tags])
        else:
            if len(self.tags) != 0:
                tag = '\n\n' + ' '.join(['#' + t for t in self.tags])
            else:
                tag = ""
        
        if mode == "text":
            mode = None 
        
        dwp = not webPreview
        
        if len(msg) <= 4096 - len(tag):
            self.tb.send_message(self.chatid, msg.strip() + tag, parse_mode=mode, disable_web_page_preview=dwp) 
        else:
            for m in telebot.util.smart_split(msg, 4096 - len(tag)):
                self.tb.send_message(self.chatid, m.strip() + tag, parse_mode=mode) 
    
if __name__ == "__main__":
    token, chatid = open("TelegramBot.ident").read().strip().split("\n")
    t = TelegramBotOfficial(token).SetChatID(int(chatid))
    # t.SendMsg(open("Telegram.py").read(), "tag1", "tag2")
    t.SendMsg("test")
    # t.SendFile("URL.py")


import js2py

#print("load " + '/'.join(__file__.split('/')[-2:]))

class JavaScript():
    def __init__(self) -> None:
        pass

    def Eval(self, code:str):
        """
        Just like javascript eval. Translates javascript to python, executes and returns python object. js is javascript source code

        EXAMPLE:

        >>> Tools.JavaScript.Eval('console.log( "Hello World!" )')
        'Hello World!'
        >>> add = Tools.JavaScript.Eval('function add(a, b) {return a + b}')
        >>> add(1, 2) + 3
        6
        >>> add('1', 2, 3)
        u'12'
        >>> add.constructor
        function Function() { [python code] }
        
        NOTE: For Js Number, String, Boolean and other base types returns appropriate python BUILTIN type. For Js functions and objects, returns Python wrapper - basically behaves like normal python object. If you really want to convert object to python dict you can use to_dict method.
        
        :param code: The code to be evaluated
        :type code: str
        :return: The result of the code being evaluated.
        """
        return js2py.eval_js(code)
    
    def Eval6(self, code:str):
        """
        Just like Eval() but with experimental support for js6 via babel.
        
        :param code: The code to be executed
        :type code: str
        :return: The return value is the result of the last statement executed.
        """
        return js2py.eval_js6(code)


# -*- coding: utf-8 -*-

from .keyword_list import KeywordList


class MySQLKeywords(KeywordList):

    KEYWORDS = [
        "ADD",
        "ALL",
        "ALTER",
        "ANALYZE",
        "AND",
        "AS",
        "ASC",
        "ASENSITIVE",
        "BEFORE",
        "BETWEEN",
        "BIGINT",
        "BINARY",
        "BLOB",
        "BOTH",
        "BY",
        "CALL",
        "CASCADE",
        "CASE",
        "CHANGE",
        "CHAR",
        "CHARACTER",
        "CHECK",
        "COLLATE",
        "COLUMN",
        "CONDITION",
        "CONNECTION",
        "CONSTRAINT",
        "CONTINUE",
        "CONVERT",
        "CREATE",
        "CROSS",
        "CURRENT_DATE",
        "CURRENT_TIME",
        "CURRENT_TIMESTAMP",
        "CURRENT_USER",
        "CURSOR",
        "DATABASE",
        "DATABASES",
        "DAY_HOUR",
        "DAY_MICROSECOND",
        "DAY_MINUTE",
        "DAY_SECOND",
        "DEC",
        "DECIMAL",
        "DECLARE",
        "DEFAULT",
        "DELAYED",
        "DELETE",
        "DESC",
        "DESCRIBE",
        "DETERMINISTIC",
        "DISTINCT",
        "DISTINCTROW",
        "DIV",
        "DOUBLE",
        "DROP",
        "DUAL",
        "EACH",
        "ELSE",
        "ELSEIF",
        "ENCLOSED",
        "ESCAPED",
        "EXISTS",
        "EXIT",
        "EXPLAIN",
        "FALSE",
        "FETCH",
        "FLOAT",
        "FLOAT4",
        "FLOAT8",
        "FOR",
        "FORCE",
        "FOREIGN",
        "FROM",
        "FULLTEXT",
        "GOTO",
        "GRANT",
        "GROUP",
        "HAVING",
        "HIGH_PRIORITY",
        "HOUR_MICROSECOND",
        "HOUR_MINUTE",
        "HOUR_SECOND",
        "IF",
        "IGNORE",
        "IN",
        "INDEX",
        "INFILE",
        "INNER",
        "INOUT",
        "INSENSITIVE",
        "INSERT",
        "INT",
        "INT1",
        "INT2",
        "INT3",
        "INT4",
        "INT8",
        "INTEGER",
        "INTERVAL",
        "INTO",
        "IS",
        "ITERATE",
        "JOIN",
        "KEY",
        "KEYS",
        "KILL",
        "LABEL",
        "LEADING",
        "LEAVE",
        "LEFT",
        "LIKE",
        "LIMIT",
        "LINES",
        "LOAD",
        "LOCALTIME",
        "LOCALTIMESTAMP",
        "LOCK",
        "LONG",
        "LONGBLOB",
        "LONGTEXT",
        "LOOP",
        "LOW_PRIORITY",
        "MATCH",
        "MEDIUMBLOB",
        "MEDIUMINT",
        "MEDIUMTEXT",
        "MIDDLEINT",
        "MINUTE_MICROSECOND",
        "MINUTE_SECOND",
        "MOD",
        "MODIFIES",
        "NATURAL",
        "NOT",
        "NO_WRITE_TO_BINLOG",
        "NULL",
        "NUMERIC",
        "ON",
        "OPTIMIZE",
        "OPTION",
        "OPTIONALLY",
        "OR",
        "ORDER",
        "OUT",
        "OUTER",
        "OUTFILE",
        "PRECISION",
        "PRIMARY",
        "PROCEDURE",
        "PURGE",
        "RAID0",
        "RANGE",
        "READ",
        "READS",
        "REAL",
        "REFERENCES",
        "REGEXP",
        "RELEASE",
        "RENAME",
        "REPEAT",
        "REPLACE",
        "REQUIRE",
        "RESTRICT",
        "RETURN",
        "REVOKE",
        "RIGHT",
        "RLIKE",
        "SCHEMA",
        "SCHEMAS",
        "SECOND_MICROSECOND",
        "SELECT",
        "SENSITIVE",
        "SEPARATOR",
        "SET",
        "SHOW",
        "SMALLINT",
        "SONAME",
        "SPATIAL",
        "SPECIFIC",
        "SQL",
        "SQLEXCEPTION",
        "SQLSTATE",
        "SQLWARNING",
        "SQL_BIG_RESULT",
        "SQL_CALC_FOUND_ROWS",
        "SQL_SMALL_RESULT",
        "SSL",
        "STARTING",
        "STRAIGHT_JOIN",
        "TABLE",
        "TERMINATED",
        "THEN",
        "TINYBLOB",
        "TINYINT",
        "TINYTEXT",
        "TO",
        "TRAILING",
        "TRIGGER",
        "TRUE",
        "UNDO",
        "UNION",
        "UNIQUE",
        "UNLOCK",
        "UNSIGNED",
        "UPDATE",
        "USAGE",
        "USE",
        "USING",
        "UTC_DATE",
        "UTC_TIME",
        "UTC_TIMESTAMP",
        "VALUES",
        "VARBINARY",
        "VARCHAR",
        "VARCHARACTER",
        "VARYING",
        "WHEN",
        "WHERE",
        "WHILE",
        "WITH",
        "WRITE",
        "X509",
        "XOR",
        "YEAR_MONTH",
        "ZEROFILL",
    ]

    def get_name(self):
        return "MySQL"



# -*- coding: utf-8 -*-


class BasePaginator(object):

    _current_page_resolver = None

    def _is_valid_page_number(self, page):
        """
        Determine if the given value is a valid page number.

        :param page: The given page number
        :type page: int

        :rtype: bool
        """
        return isinstance(page, int) and page >= 1

    @property
    def items(self):
        """
        Get the slice of items being paginated.

        :rtype: list
        """
        return self._items.all()

    @property
    def first_item(self):
        """
        Get the number of the first item in the slice.

        :rtype: int
        """
        return (self.current_page - 1) * self.per_page + 1

    @property
    def last_item(self):
        """
        Get the number of the last item in the slice.

        :rtype: int
        """
        return self.first_item + self.count() - 1

    def has_pages(self):
        """
        Determine if there are enough items to split into multiple pages.

        :rtype: int
        """
        return not (self.current_page == 1 and not self.has_more_pages())

    def is_empty(self):
        """
        Determine if the list of items is empty or not.

        :rtype: bool
        """
        return self._items.is_empty()

    def count(self):
        """
        Get the number of items for the current page.

        :rtype: int
        """
        return len(self._items)

    @property
    def previous_page(self):
        if self.current_page > 1:
            return self.current_page - 1

    @property
    def next_page(self):
        if self.has_more_pages():
            return self.current_page + 1

    def get_collection(self):
        return self._items

    @classmethod
    def resolve_current_page(cls, default=1):
        if cls._current_page_resolver is not None:
            return cls._current_page_resolver()

        return default

    @classmethod
    def current_page_resolver(cls, resolver):
        cls._current_page_resolver = staticmethod(resolver)

    def __len__(self):
        return self.count()

    def __iter__(self):
        for item in self._items:
            yield item

    def __getitem__(self, item):
        return self.items[item]



from mitmproxy.tools.main import mitmdump
from .. import Process

class Mitmproxy():
    def __init__(
            self, 
            addonFilePath:str|list[str]=None, 
            saveStreamToFile:str=None,
            listenAddress:str="0.0.0.0",
            listenPort:int=8080,
            mode:str="http", # socks5
            sslInsecure:bool=False,
        ) -> None:

        args = [
            "--quiet",
        ]

        if addonFilePath != None:
            if type(addonFilePath) == str:
                args.append("--scripts")
                args.append(addonFilePath)
            elif type(addonFilePath) == list:
                for a in addonFilePath:
                    args.append("--scripts")
                    args.append(addonFilePath)
        
        if saveStreamToFile != None:
            args.append('--save-stream-file')
            args.append(f"+{saveStreamToFile}")
        
        if mode == 'http':
            mode = 'regular'

        args.append("--listen-host")
        args.append(listenAddress)

        args.append("--listen-port")
        args.append(str(listenPort))
        
        args.append("--mode")
        args.append(mode)

        if sslInsecure == False:
            args.append('--ssl-insecure')

        try:
            self.pm = Process(mitmdump, args)
        except RuntimeError:
            raise Exception('需要在__name__ == "__main__"中执行Mitmproxy')

    def Stop(self):
        self.pm.Kill()



# from bagbag import Tools, String, Range, Funcs, Hash, Os, Lg
import typing 

#print("load " + '/'.join(__file__.split('/')[-2:]))

try:
    from .. import Tools
    from .. import String
    from ..Python import Range 
    from .. import Funcs 
    from .. import Hash 
    from .. import Os 
    from .. import Lg
except:
    import sys 
    sys.path.append("..")
    import Tools
    import String 
    from Python import Range 
    import Funcs 
    import Hash 
    import Os
    import Lg

class WebCrawlerResult:
    def __init__(self, URL:str="", PageSource:str="", Title:str=""):
        self.URL = URL 
        self.PageSource = PageSource
        self.Title = Title 
    
    def __repr__(self) -> str:
        pg = String(self.PageSource).Ommit(180).replace("\n", "\\n").replace("\t", "\\t")
        return f"WebCrawlerResult(URL={self.URL} Title={self.Title}) PageSource={pg}"
    
    def __str__(self) -> str:
        return self.__repr__()

class WebCrawler():
    def __init__(self):
        self.cachedbfname = Funcs.UUID()
        self.db = Tools.SQLite(self.cachedbfname)
        (
            self.db.Table("queue"). 
                AddColumn("url", "text"). 
                AddColumn("fetched", "int").
                AddColumn("md5", "string"). 
                AddIndex("fetched"). 
                AddIndex("md5")
        )
        self.tb = self.db.Table("queue")
    
    def getPageSource(self, se:Tools.Selenium.Chrome, url:str) -> list[str]:
        content = None 
        title = None 
        for _ in Range(3):
            try:
                se.Get(url)
                content = se.PageSource()
                title = se.Title()
                return title, content 
            except KeyboardInterrupt:
                # Lg.Trace("用户中断, 清理, 退出.")
                self.Close()
            except:
                pass 
        return title, content  
    
    def Run(self, url:str) -> typing.Iterable[WebCrawlerResult]:
        self.tb.Data({
            "url": url,
            "fetched": 0,
            "md5": Hash.Md5sum(url)
        }).Insert()

        with Tools.Selenium.Chrome(randomUA=False) as se:
            while self.tb.Where("fetched", "=", 0).Count() != 0:
                row = self.tb.Where("fetched", "=", 0).First()

                # Lg.Trace(row)
                nexturl = row['url']

                # Time.Sleep(5)
                # ipdb.set_trace()

                # Lg.Trace("采集URL:", nexturl)

                try:
                    u = Tools.URL(nexturl).Parse()
                except KeyboardInterrupt:
                    # Lg.Trace("用户中断, 清理, 退出.")
                    self.Close()
                except Exception as e:
                    # Lg.Trace(nexturl)
                    # Lg.Trace(traceback.print_exc())
                    continue

                schema = u.Schema
                host = u.Host
                port = u.Port
                path = Os.Path.Basedir(u.Path)

                # Lg.Trace("域名:", host)

                # Lg.Trace(host, "获取页面:", nexturl)
                title, content = self.getPageSource(se, nexturl)
                if not content:
                    # Lg.Trace(host, "页面获取失败")
                    continue 

                # Lg.Trace(host, "页面获取完成")

                yield WebCrawlerResult(
                    URL=nexturl,
                    PageSource=content,
                    Title=title,
                )

                self.tb.Where("id", "=", row['id']).Data({
                    "fetched": 1
                }).Update()

                for i in String(content).RegexFind("<a.+?href=\"(.*?)\".*?>(.+?)</a>", True):
                    # Lg.Trace("正则找到的结果:", i)
                    try:
                        curl = i[1].strip()
                        if "#" in curl:
                            curl = curl.split("#")[0]
                            # Lg.Trace("URL有#, 去掉之后:", curl)
                        
                        if curl.strip() == "":
                            continue

                        if '/../' in curl:
                            continue
                        
                        # Lg.Trace("找到链接:", curl)
                        if len(String(curl).RegexFind("(.+)://.+")) != 0:
                            # Lg.Trace("链接是完整链接")
                            if String(curl).RegexFind("(.+)://.+")[0][1] in ["http", "https"]:
                                if Tools.URL(curl).Parse().Host != host:
                                    # Lg.Trace("链接是站外链, 跳过")
                                    continue 
                                else:
                                    if self.tb.Where("md5", "=", Hash.Md5sum(curl)).Exists():
                                        # Lg.Trace("链接已爬过/在队列, 跳过")
                                        continue
                                    
                                    # Lg.Trace("放入队列:", curl)
                                    self.tb.Data({
                                        "url": curl,
                                        "fetched": 0,
                                        "md5": Hash.Md5sum(curl)
                                    }).Insert()
                            else:
                                # Lg.Trace("链接不是web的协议, 跳过")
                                pass
                        else:
                            if curl.startswith("#"):
                                # Lg.Trace("链接是tag, 跳过")
                                continue
                            
                            if curl.startswith("javascript:"):
                                # Lg.Trace("链接是javascript, 跳过")
                                continue
                            
                            if curl.startswith("mailto:"):
                                # Lg.Trace("链接是mailto, 跳过")
                                continue
                            
                            if (schema == "http" and port == "80") or (schema == "https" and port == "443"):
                                nnexturl = schema + "://" + host
                            else:
                                nnexturl = schema + "://" + host + ":" + str(port)
                            
                            if curl.startswith("/"):
                                # Lg.Trace("链接是绝对路径")
                                nnexturl += curl
                            else:
                                # Lg.Trace("链接是相对路径")
                                nnexturl += Os.Path.Join(path, curl)
 
                            if self.tb.Where("md5", "=", Hash.Md5sum(nnexturl)).Exists():
                                # Lg.Trace("链接已爬过/在队列, 跳过")
                                continue

                            # Lg.Trace("放入队列:", nnexturl)
                            self.tb.Data({
                                "url": nnexturl,
                                "fetched": 0,
                                "md5": Hash.Md5sum(nnexturl)
                            }).Insert()

                    except KeyboardInterrupt:
                        # Lg.Trace("用户中断, 清理, 退出.")
                        self.Close()
                    except Exception as e:
                        Lg.Error("解析链接\""+i[1]+"\"出错:", e)

                # Lg.Trace("解析页面完成")

    def Close(self):
        self.db.Close()
        Os.Unlink(self.cachedbfname) 

    def __enter__(self):
        return self 
    
    def __exit__(self, exc_type, exc_value, traceback):
        try:
            self.Close()
        except:
            pass

if __name__ == "__main__":
    # Lg.SetFile("230412.web.crawler.log", color=False)
    with WebCrawler() as wc:
        for r in wc.Run('https://www.spaceblock.co.uk'):
            Lg.Info(r.URL, "=====>", r.Title)





# -*- coding: utf-8 -*-

import types
from functools import update_wrapper
from .relations.wrapper import Wrapper
from .builder import Builder
from ..query import QueryBuilder
from .relations import (
    HasOne,
    HasMany,
    HasManyThrough,
    BelongsTo,
    BelongsToMany,
    MorphOne,
    MorphMany,
    MorphTo,
    MorphToMany,
)


class accessor(object):
    def __init__(self, accessor_, attribute=None):
        self.accessor = accessor_
        self.mutator_ = None
        if attribute is not None:
            self.attribute = attribute
        else:
            if isinstance(accessor_, property):
                self.attribute = accessor_.fget.__name__
            else:
                self.attribute = self.accessor.__name__

        self.expr = accessor_
        if accessor_ is not None:
            update_wrapper(self, accessor_)

    def __get__(self, instance, owner):
        if instance is None:
            return self.expr
        else:
            return self.accessor(instance)

    def __set__(self, instance, value):
        if self.mutator_ is None:
            return instance.set_attribute(self.attribute, value)

        self.mutator_(instance, value)

    def mutator(self, f):
        self.mutator_ = f

        return mutator(f, self.attribute)


class mutator(object):
    def __init__(self, mutator_, attribute=None):
        self.mutator = mutator_
        self.accessor_ = None
        self.attribute = attribute or self.mutator.__name__

    def __get__(self, instance, owner):
        if instance is None:
            return self.mutator
        else:
            if self.accessor_ is None:
                return instance.get_attribute(self.attribute)

            return self.accessor_(instance)

    def __set__(self, instance, value):
        self.mutator(instance, value)

    def accessor(self, f):
        self.accessor_ = f

        return accessor(f, self.attribute)


class column(object):
    def __init__(self, property_, attribute=None):
        self.property = property_
        self.mutator_ = None
        self.accessor_ = None
        if attribute is not None:
            self.attribute = attribute
        else:
            if isinstance(property_, property):
                self.attribute = property_.fget.__name__
            else:
                self.attribute = self.property.__name__

    def __get__(self, instance, owner):
        if instance is None:
            return self.mutator_
        else:
            if self.accessor_ is None:
                return instance.get_attribute(self.attribute)

            return self.accessor_(instance)

    def __set__(self, instance, value):
        if self.mutator_ is None:
            return instance.set_attribute(self.attribute, value)

        self.mutator_(instance, value)

    def mutator(self, f):
        self.mutator_ = f

        return mutator(f, self.attribute)

    def accessor(self, f):
        self.accessor_ = f

        return accessor(f, self.attribute)


class scope(classmethod):
    """
    Decorator to add local scopes.
    """

    def __init__(self, method):
        super(scope, self).__init__(method)

        self._method = method
        self._owner = None

        update_wrapper(self, method)

    def __get__(self, instance, owner, *args, **kwargs):
        if instance:
            self._owner = None
        else:
            self._owner = owner

        return self

    def __call__(self, *args, **kwargs):
        if not self._owner:
            return self._method(self._owner, *args, **kwargs)
        else:
            return getattr(self._owner.query(), self._method.__name__)(*args, **kwargs)


# Relations decorators
class relation(object):
    """
    Base relation decorator
    """

    relation_class = None

    def __init__(self, func=None, relation=None):
        self._relation = relation
        self._related = None
        self._conditions = None

        self.set_func(func)

    def set_func(self, func):
        self.func = func

        if self._relation is None:
            if isinstance(func, property):
                self._relation = func.fget.__name__ if func else None
            else:
                self._relation = func.__name__ if func else None

        self.expr = func
        if func is not None:
            update_wrapper(self, func)

    def __get__(self, instance, owner):
        if instance is None:
            return self.expr

        if self._relation in instance._relations:
            return instance._relations[self._relation]

        self._related = self.func(instance)
        if isinstance(self._related, (Builder, QueryBuilder)):
            # Extra conditions on relation
            self._conditions = self._related
            self._related = self._related.get_model().__class__

        relation = self._get(instance)

        if self._conditions:
            # Setting extra conditions
            self._set_conditions(relation)

        relation = Wrapper(relation)

        instance._relations[self._relation] = relation

        return relation

    def _get(self, instance):
        raise NotImplementedError()

    def _set_conditions(self, relation):
        relation.merge_query(self._conditions)
        relation.set_extra_query(self._conditions)

    def __call__(self, func):
        self.set_func(func)

        return self


class has_one(relation):
    """
    Has One relationship decorator
    """

    relation_class = HasOne

    def __init__(self, foreign_key=None, local_key=None, relation=None):
        if isinstance(foreign_key, (types.FunctionType, types.MethodType)):
            func = foreign_key
            foreign_key = None
        else:
            func = None

        self._foreign_key = foreign_key
        self._local_key = local_key

        super(has_one, self).__init__(func, relation)

    def _get(self, instance):
        return instance.has_one(
            self._related,
            self._foreign_key,
            self._local_key,
            self._relation,
            _wrapped=False,
        )


class morph_one(relation):
    """
    Morph One relationship decorator
    """

    relation_class = MorphOne

    def __init__(
        self, name, type_column=None, id_column=None, local_key=None, relation=None
    ):
        if isinstance(name, (types.FunctionType, types.MethodType)):
            raise RuntimeError("morph_one relation requires a name")

        self._name = name
        self._type_column = type_column
        self._id_column = id_column
        self._local_key = local_key

        super(morph_one, self).__init__(relation=relation)

    def _get(self, instance):
        return instance.morph_one(
            self._related,
            self._name,
            self._type_column,
            self._id_column,
            self._local_key,
            self._relation,
            _wrapped=False,
        )


class belongs_to(relation):
    """
    Belongs to relationship decorator
    """

    relation_class = BelongsTo

    def __init__(self, foreign_key=None, other_key=None, relation=None):
        if isinstance(foreign_key, (types.FunctionType, types.MethodType)):
            func = foreign_key
            foreign_key = None
        else:
            func = None

        self._foreign_key = foreign_key
        self._other_key = other_key

        super(belongs_to, self).__init__(func, relation)

    def _get(self, instance):
        return instance.belongs_to(
            self._related,
            self._foreign_key,
            self._other_key,
            self._relation,
            _wrapped=False,
        )

    def _set(self, relation):
        relation._foreign_key = self._foreign_key
        relation._other_key = self._other_key
        relation._relation = self._relation


class morph_to(relation):
    """
    Morph To relationship decorator
    """

    relation_class = MorphTo

    def __init__(self, name=None, type_column=None, id_column=None):
        if isinstance(name, (types.FunctionType, types.MethodType)):
            func = name
            name = None
        else:
            func = None

        self._name = name
        self._type_column = type_column
        self._id_column = id_column

        super(morph_to, self).__init__(func, name)

    def _get(self, instance):
        return instance.morph_to(
            self._relation, self._type_column, self._id_column, _wrapped=False
        )


class has_many(relation):
    """
    Has Many relationship decorator
    """

    relation_class = HasMany

    def __init__(self, foreign_key=None, local_key=None, relation=None):
        if isinstance(foreign_key, (types.FunctionType, types.MethodType)):
            func = foreign_key
            foreign_key = None
        else:
            func = None

        self._foreign_key = foreign_key
        self._local_key = local_key

        super(has_many, self).__init__(func, relation)

    def _get(self, instance):
        return instance.has_many(
            self._related,
            self._foreign_key,
            self._local_key,
            self._relation,
            _wrapped=False,
        )


class has_many_through(relation):
    """
    Has Many Through relationship decorator
    """

    relation_class = HasManyThrough

    def __init__(self, through, first_key=None, second_key=None, relation=None):
        if isinstance(through, (types.FunctionType, types.MethodType)):
            raise RuntimeError(
                "has_many_through relation requires the through parameter"
            )

        self._through = through
        self._first_key = first_key
        self._second_key = second_key

        super(has_many_through, self).__init__(relation=relation)

    def _get(self, instance):
        return instance.has_many_through(
            self._related,
            self._through,
            self._first_key,
            self._second_key,
            self._relation,
            _wrapped=False,
        )


class morph_many(relation):
    """
    Morph Many relationship decorator
    """

    relation_class = MorphMany

    def __init__(
        self, name, type_column=None, id_column=None, local_key=None, relation=None
    ):
        if isinstance(name, (types.FunctionType, types.MethodType)):
            raise RuntimeError("morph_many relation requires a name")

        self._name = name
        self._type_column = type_column
        self._id_column = id_column
        self._local_key = local_key

        super(morph_many, self).__init__(relation=relation)

    def _get(self, instance):
        return instance.morph_many(
            self._related,
            self._name,
            self._type_column,
            self._id_column,
            self._local_key,
            self._relation,
            _wrapped=False,
        )


class belongs_to_many(relation):
    """
    Belongs To Many relationship decorator
    """

    relation_class = BelongsToMany

    def __init__(
        self,
        table=None,
        foreign_key=None,
        other_key=None,
        relation=None,
        with_timestamps=False,
        with_pivot=None,
    ):
        if isinstance(table, (types.FunctionType, types.MethodType)):
            func = table
            table = None
        else:
            func = None

        self._table = table
        self._foreign_key = foreign_key
        self._other_key = other_key

        self._timestamps = with_timestamps
        self._pivot = with_pivot

        super(belongs_to_many, self).__init__(func, relation)

    def _get(self, instance):
        r = instance.belongs_to_many(
            self._related,
            self._table,
            self._foreign_key,
            self._other_key,
            self._relation,
            _wrapped=False,
        )

        if self._timestamps:
            r = r.with_timestamps()

        if self._pivot:
            r = r.with_pivot(*self._pivot)

        return r


class morph_to_many(relation):
    """
    Morph To Many relationship decorator
    """

    relation_class = MorphToMany

    def __init__(
        self, name, table=None, foreign_key=None, other_key=None, relation=None
    ):
        if isinstance(name, (types.FunctionType, types.MethodType)):
            raise RuntimeError("morph_to_many relation required a name")

        self._name = name
        self._table = table
        self._foreign_key = foreign_key
        self._other_key = other_key

        super(morph_to_many, self).__init__(relation=relation)

    def _get(self, instance):
        return instance.morph_to_many(
            self._related,
            self._name,
            self._table,
            self._foreign_key,
            self._other_key,
            relation=self._relation,
            _wrapped=False,
        )


class morphed_by_many(relation):
    """
    Morphed By Many relationship decorator
    """

    relation_class = MorphToMany

    def __init__(
        self, name, table=None, foreign_key=None, other_key=None, relation=None
    ):
        if isinstance(foreign_key, (types.FunctionType, types.MethodType)):
            raise RuntimeError("morphed_by_many relation requires a name")

        self._name = name
        self._table = table
        self._foreign_key = foreign_key
        self._other_key = other_key

        super(morphed_by_many, self).__init__(relation=relation)

    def _get(self, instance):
        return instance.morphed_by_many(
            self._related,
            self._name,
            self._table,
            self._foreign_key,
            self._other_key,
            self._relation,
            _wrapped=False,
        )



# from .Elevated import Elevated
# from .Essential import Essential
# from .Elevated import twitterTweet as ElevatedTweet

from typing import TYPE_CHECKING
from lazy_imports import LazyImporter
import sys

_import_structure = {
    "Elevated_src": [
        "Elevated",
        "ElevatedTweet",
        "ElevatedTwitterUser"
    ], 
    "Essential_src": [
        "Essential"
    ], 
    "Browser_src": [
        "Browser",
        "BrowserTwitterUser"
    ],
    "Utils": [
        "Utils"
    ],
    "Nitter_src": [
        "Nitter",
        "NitterTweet", 
        "NitterTwitterUser"
    ]
}

if TYPE_CHECKING:
    from .Elevated_src import (
        Elevated,
        ElevatedTweet,
        ElevatedTwitterUser,
    )
    from .Essential_src import Essential
    from .Browser_src import Browser, BrowserTwitterUser
    from .Nitter_src import Nitter, NitterTweet, NitterTwitterUser
    from . import Utils
else:
    sys.modules[__name__] = LazyImporter(
        __name__,
        globals()["__file__"],
        _import_structure,
        extra_objects={},
    )





from __future__ import annotations

from .. import Http
from .. import Lg
from .. import Random
from .. import Hash
from .. import Json
from ..Http import useragents 

#print("load " + '/'.join(__file__.split('/')[-2:]))

from . import pygtrans

class Baidu():
    def __init__(self, appid:str, secretkey:str) -> None:
        self.appid = str(appid)
        self.secretkey = str(secretkey)
        self.apiurl = "http://api.fanyi.baidu.com/api/trans/vip/translate"
        self.to = "zh"
        self.From = "auto"

        self.SetLang()

    def SetLang(self, To:str="zh", From:str="auto") -> Baidu:
        self.ffrom = From 
        self.to = To 
        return self 
    
    def Translate(self, text:str) -> str:
        if "\n" in text:
            raise Exception("不允许换行符哦")

        salt = str(Random.Int(32768, 65536))
        preSign = self.appid + text + salt + self.secretkey
        sign = Hash.Md5sum(preSign)
        params = {
            "q":     text,
            "from":  self.ffrom,
            "to":    self.to,
            "appid": self.appid,
            "salt":  salt,
            "sign":  sign,
        }
        resp = Http.Get(self.apiurl, params)
        if resp.StatusCode != 200:
            raise Exception(f"翻译出错, 状态码: {resp.StatusCode}, 返回内容: {resp.Content}")
        
        try:    
            rj = Json.Loads(resp.Content)['trans_result'][0]['dst']
        except:
            raise Exception(f"翻译出错:{resp.Content}")

        return rj 

class Google():
    def __init__(self, httpProxy:str=None, retryTimes:int=3) -> None:
        self.httpProxy = httpProxy
        self.to = "zh-CN"
        self.From = "auto"
        self.retryTimes = retryTimes

    def SetLang(self, To:str="zh-CN", From:str="auto") -> Google:
        self.From = From 
        self.to = To 
        return self 
    
    def Translate(self, text:str, format:str="text") -> str:
        """
        It translates the text from one language to another.
        
        :param text: The text to be translated
        :type text: str
        :param format: The format of the text to be translated, defaults to html. 可选html或者text
        :type format: str (optional)
        """

        errtimes = 0
        while True:
            try:
                client = pygtrans.Translate(
                    target=self.to,
                    source=self.From,
                    fmt='html',
                    user_agent=Random.Choice(useragents)['user_agent'],
                    domain='com', # cn或者com
                    proxies=None if self.httpProxy == None else {"http": self.httpProxy, "https": self.httpProxy},
                )
                return client.translate(text).translatedText
            except Exception as e:
                Lg.Warn("翻译出错:", e)
                errtimes += 1

                if errtimes > self.retryTimes:
                    raise e


class NLLB():
    # version: '3'
    # services:
    #   nllb-arm64-facebook-nllb-200-distilled-600m:
    #     image: darren2046/nllb-arm64-facebook-nllb-200-distilled-600m
    #     container_name: nllb-arm64-facebook-nllb-200-distilled-600m-192.168.168.77
    #     restart: always
    #     ports:
    #       - 6060:6060

    def __init__(self, server:str) -> None:
        self.server = server 
        self.From = "eng_Latr"
        self.To = "zho_Hans"

        if not self.server.startswith("http://") and not self.server.startswith("https://"):
            self.server = "https://" + self.server + "/translate"
        if not self.server.endswith("/translate"):
            self.server = self.server + "/translate"
    
    def SetLang(self, To:str="zho_Hans", From:str="eng_Latr"):
        self.To = To 
        self.From = From
    
    def Translate(self, text:str) -> str:
        res = Http.PostForm(self.server, {
            "src_lang": self.From,
            "tgt_lang": self.To, 
            "source": text
        }, timeout=180, timeoutRetryTimes=3)
        # Lg.Trace(res.Content)

        try:
            res = Json.Loads(res.Content)
            res["translation"][0]
        except Exception as e:
            Lg.Trace(res)
            Lg.Error("载入返回内容错误")
            raise e
        
        return res["translation"][0]

# class Translater:
#     NLLB
#     Google
#     Baidu

if __name__ == "__main__":
    # appid, secretkey = open("baidu.ident").read().strip().split(',')
    # b = Baidu(appid, secretkey).SetLang("zh", "auto")
    # text = b.Translate("This is a test")
    # Lg.Trace(text)

    # g = Google("http://192.168.1.186:8899").SetLang("zh-CN")
    # text = g.Translate("This is a test")
    # Lg.Trace(text)

    n = NLLB("example.com")
    text = n.Translate("No Language Left Behind")
    Lg.Trace(text)


# -*- coding: utf-8 -*-

import re


class Qmarker(object):

    RE_QMARK = re.compile(r"\?\?|\?|%")

    @classmethod
    def qmark(cls, query):
        """
        Convert a "qmark" query into "format" style.
        """

        def sub_sequence(m):
            s = m.group(0)
            if s == "??":
                return "?"
            if s == "%":
                return "%%"
            else:
                return "%s"

        return cls.RE_QMARK.sub(sub_sequence, query)

    @classmethod
    def denullify(cls, args):
        for arg in args:
            if arg is not None:
                yield arg
            else:
                yield ()


qmark = Qmarker.qmark
denullify = Qmarker.denullify



# -*- coding: utf-8 -*-

from .platform import Platform
from .keywords.mysql_keywords import MySQLKeywords
from ..identifier import Identifier


class MySQLPlatform(Platform):

    LENGTH_LIMIT_TINYTEXT = 255
    LENGTH_LIMIT_TEXT = 65535
    LENGTH_LIMIT_MEDIUMTEXT = 16777215

    LENGTH_LIMIT_TINYBLOB = 255
    LENGTH_LIMIT_BLOB = 65535
    LENGTH_LIMIT_MEDIUMBLOB = 16777215

    INTERNAL_TYPE_MAPPING = {
        "tinyint": "boolean",
        "smallint": "smallint",
        "mediumint": "integer",
        "int": "integer",
        "integer": "integer",
        "bigint": "bigint",
        "int8": "bigint",
        "bool": "boolean",
        "boolean": "boolean",
        "tinytext": "text",
        "mediumtext": "text",
        "longtext": "text",
        "text": "text",
        "varchar": "string",
        "string": "string",
        "char": "string",
        "date": "date",
        "datetime": "datetime",
        "timestamp": "datetime",
        "time": "time",
        "float": "float",
        "double": "float",
        "real": "float",
        "decimal": "decimal",
        "numeric": "decimal",
        "year": "date",
        "longblob": "blob",
        "blob": "blob",
        "mediumblob": "blob",
        "tinyblob": "blob",
        "binary": "binary",
        "varbinary": "binary",
        "set": "simple_array",
        "enum": "enum",
    }

    def get_list_table_columns_sql(self, table, database=None):
        if database:
            database = "'%s'" % database
        else:
            database = "DATABASE()"

        return (
            "SELECT COLUMN_NAME AS field, COLUMN_TYPE AS type, IS_NULLABLE AS `null`, "
            "COLUMN_KEY AS `key`, COLUMN_DEFAULT AS `default`, EXTRA AS extra, COLUMN_COMMENT AS comment, "
            "CHARACTER_SET_NAME AS character_set, COLLATION_NAME AS collation "
            "FROM information_schema.COLUMNS WHERE TABLE_SCHEMA = %s AND TABLE_NAME = '%s'"
            % (database, table)
        )

    def get_list_table_indexes_sql(self, table, current_database=None):
        sql = """
            SELECT TABLE_NAME AS `Table`, NON_UNIQUE AS Non_Unique, INDEX_NAME AS Key_name,
            SEQ_IN_INDEX AS Seq_in_index, COLUMN_NAME AS Column_Name, COLLATION AS Collation,
            CARDINALITY AS Cardinality, SUB_PART AS Sub_Part, PACKED AS Packed,
            NULLABLE AS `Null`, INDEX_TYPE AS Index_Type, COMMENT AS Comment
            FROM information_schema.STATISTICS WHERE TABLE_NAME = '%s'
        """

        if current_database:
            sql += " AND TABLE_SCHEMA = '%s'" % current_database

        return sql % table

    def get_list_table_foreign_keys_sql(self, table, database=None):
        sql = (
            "SELECT DISTINCT k.`CONSTRAINT_NAME`, k.`COLUMN_NAME`, k.`REFERENCED_TABLE_NAME`, "
            "k.`REFERENCED_COLUMN_NAME` /*!50116 , c.update_rule, c.delete_rule */ "
            "FROM information_schema.key_column_usage k /*!50116 "
            "INNER JOIN information_schema.referential_constraints c ON "
            "  c.constraint_name = k.constraint_name AND "
            "  c.table_name = '%s' */ WHERE k.table_name = '%s'" % (table, table)
        )

        if database:
            sql += (
                " AND k.table_schema = '%s' /*!50116 AND c.constraint_schema = '%s' */"
                % (database, database)
            )

        sql += " AND k.`REFERENCED_COLUMN_NAME` IS NOT NULL"

        return sql

    def get_alter_table_sql(self, diff):
        """
        Get the ALTER TABLE SQL statement

        :param diff: The table diff
        :type diff: orator.dbal.table_diff.TableDiff

        :rtype: list
        """
        column_sql = []
        query_parts = []

        if diff.new_name is not False:
            query_parts.append(
                "RENAME TO %s" % diff.get_new_name().get_quoted_name(self)
            )

        # Added columns?

        # Removed columns?

        for column_diff in diff.changed_columns.values():
            column = column_diff.column
            column_dict = column.to_dict()

            # Don't propagate default value changes for unsupported column types.
            if (
                column_diff.has_changed("default")
                and len(column_diff.changed_properties) == 1
                and (column_dict["type"] == "text" or column_dict["type"] == "blob")
            ):
                continue

            query_parts.append(
                "CHANGE %s %s"
                % (
                    column_diff.get_old_column_name().get_quoted_name(self),
                    self.get_column_declaration_sql(
                        column.get_quoted_name(self), column_dict
                    ),
                )
            )

        for old_column_name, column in diff.renamed_columns.items():
            column_dict = column.to_dict()
            old_column_name = Identifier(old_column_name)
            query_parts.append(
                "CHANGE %s %s"
                % (
                    self.quote(old_column_name.get_quoted_name(self)),
                    self.get_column_declaration_sql(
                        self.quote(column.get_quoted_name(self)), column_dict
                    ),
                )
            )

        sql = []

        if len(query_parts) > 0:
            sql.append(
                "ALTER TABLE %s %s"
                % (diff.get_name(self).get_quoted_name(self), ", ".join(query_parts))
            )

        return sql

    def convert_booleans(self, item):
        if isinstance(item, list):
            for i, value in enumerate(item):
                if isinstance(value, bool):
                    item[i] = str(value).lower()
        elif isinstance(item, bool):
            item = str(item).lower()

        return item

    def get_boolean_type_declaration_sql(self, column):
        return "TINYINT(1)"

    def get_integer_type_declaration_sql(self, column):
        return "INT " + self._get_common_integer_type_declaration_sql(column)

    def get_bigint_type_declaration_sql(self, column):
        return "BIGINT " + self._get_common_integer_type_declaration_sql(column)

    def get_smallint_type_declaration_sql(self, column):
        return "SMALLINT " + self._get_common_integer_type_declaration_sql(column)

    def get_guid_type_declaration_sql(self, column):
        return "UUID"

    def get_datetime_type_declaration_sql(self, column):
        if "version" in column and column["version"] == True:
            return "TIMESTAMP"

        return "DATETIME"

    def get_date_type_declaration_sql(self, column):
        return "DATE"

    def get_time_type_declaration_sql(self, column):
        return "TIME"

    def get_varchar_type_declaration_sql_snippet(self, length, fixed):
        if fixed:
            return "CHAR(%s)" % length if length else "CHAR(255)"
        else:
            return "VARCHAR(%s)" % length if length else "VARCHAR(255)"

    def get_binary_type_declaration_sql_snippet(self, length, fixed):
        if fixed:
            return "BINARY(%s)" % (length or 255)
        else:
            return "VARBINARY(%s)" % (length or 255)

    def get_text_type_declaration_sql(self, column):
        length = column.get("length")
        if length:
            if length <= self.LENGTH_LIMIT_TINYTEXT:
                return "TINYTEXT"

            if length <= self.LENGTH_LIMIT_TEXT:
                return "TEXT"

            if length <= self.LENGTH_LIMIT_MEDIUMTEXT:
                return "MEDIUMTEXT"

        return "LONGTEXT"

    def get_blob_type_declaration_sql(self, column):
        length = column.get("length")
        if length:
            if length <= self.LENGTH_LIMIT_TINYBLOB:
                return "TINYBLOB"

            if length <= self.LENGTH_LIMIT_BLOB:
                return "BLOB"

            if length <= self.LENGTH_LIMIT_MEDIUMBLOB:
                return "MEDIUMBLOB"

        return "LONGBLOB"

    def get_clob_type_declaration_sql(self, column):
        length = column.get("length")
        if length:
            if length <= self.LENGTH_LIMIT_TINYTEXT:
                return "TINYTEXT"

            if length <= self.LENGTH_LIMIT_TEXT:
                return "TEXT"

            if length <= self.LENGTH_LIMIT_MEDIUMTEXT:
                return "MEDIUMTEXT"

        return "LONGTEXT"

    def get_decimal_type_declaration_sql(self, column):
        decl = super(MySQLPlatform, self).get_decimal_type_declaration_sql(column)

        return decl + self.get_unsigned_declaration(column)

    def get_unsigned_declaration(self, column):
        if column.get("unsigned"):
            return " UNSIGNED"

        return ""

    def _get_common_integer_type_declaration_sql(self, column):
        autoinc = ""
        if column.get("autoincrement"):
            autoinc = " AUTO_INCREMENT"

        return self.get_unsigned_declaration(column) + autoinc

    def get_float_type_declaration_sql(self, column):
        return "DOUBLE PRECISION" + self.get_unsigned_declaration(column)

    def get_enum_type_declaration_sql(self, column):
        return "ENUM{}".format(column["extra"]["definition"])

    def supports_foreign_key_constraints(self):
        return True

    def supports_column_collation(self):
        return False

    def quote(self, name):
        return "`%s`" % name.replace("`", "``")

    def _get_reserved_keywords_class(self):
        return MySQLKeywords

    def get_identifier_quote_character(self):
        return "`"



# -*- coding: utf-8 -*-

from .grammar import SchemaGrammar
from ..blueprint import Blueprint
from ...query.expression import QueryExpression
from ...support.fluent import Fluent


class PostgresSchemaGrammar(SchemaGrammar):

    _modifiers = ["increment", "nullable", "default"]

    _serials = [
        "big_integer",
        "integer",
        "medium_integer",
        "small_integer",
        "tiny_integer",
    ]

    marker = "%s"

    def compile_rename_column(self, blueprint, command, connection):
        """
        Compile a rename column command.

        :param blueprint: The blueprint
        :type blueprint: Blueprint

        :param command: The command
        :type command: Fluent

        :param connection: The connection
        :type connection: orator.connections.Connection

        :rtype: list
        """
        table = self.get_table_prefix() + blueprint.get_table()

        column = self.wrap(command.from_)

        return "ALTER TABLE %s RENAME COLUMN %s TO %s" % (
            table,
            column,
            self.wrap(command.to),
        )

    def compile_table_exists(self):
        """
        Compile the query to determine if a table exists

        :rtype: str
        """
        return (
            "SELECT * "
            "FROM information_schema.tables "
            "WHERE table_name = %(marker)s" % {"marker": self.get_marker()}
        )

    def compile_column_exists(self, table):
        """
        Compile the query to determine the list of columns.
        """
        return (
            "SELECT column_name "
            "FROM information_schema.columns "
            "WHERE table_name = '%s'" % table
        )

    def compile_create(self, blueprint, command, _):
        """
        Compile a create table command.
        """
        columns = ", ".join(self._get_columns(blueprint))

        return "CREATE TABLE %s (%s)" % (self.wrap_table(blueprint), columns)

    def compile_add(self, blueprint, command, _):
        table = self.wrap_table(blueprint)

        columns = self.prefix_list("ADD COLUMN", self._get_columns(blueprint))

        return "ALTER TABLE %s %s" % (table, ", ".join(columns))

    def compile_primary(self, blueprint, command, _):
        columns = self.columnize(command.columns)

        return "ALTER TABLE %s ADD PRIMARY KEY (%s)" % (
            self.wrap_table(blueprint),
            columns,
        )

    def compile_unique(self, blueprint, command, _):
        columns = self.columnize(command.columns)

        table = self.wrap_table(blueprint)

        return "ALTER TABLE %s ADD CONSTRAINT %s UNIQUE (%s)" % (
            table,
            command.index,
            columns,
        )

    def compile_index(self, blueprint, command, _):
        columns = self.columnize(command.columns)

        table = self.wrap_table(blueprint)

        return "CREATE INDEX %s ON %s (%s)" % (command.index, table, columns)

    def compile_drop(self, blueprint, command, _):
        return "DROP TABLE %s" % self.wrap_table(blueprint)

    def compile_drop_if_exists(self, blueprint, command, _):
        return "DROP TABLE IF EXISTS %s" % self.wrap_table(blueprint)

    def compile_drop_column(self, blueprint, command, connection):
        columns = self.prefix_list("DROP COLUMN", self.wrap_list(command.columns))

        table = self.wrap_table(blueprint)

        return "ALTER TABLE %s %s" % (table, ", ".join(columns))

    def compile_drop_primary(self, blueprint, command, _):
        table = blueprint.get_table()

        return "ALTER TABLE %s DROP CONSTRAINT %s_pkey" % (
            self.wrap_table(blueprint),
            table,
        )

    def compile_drop_unique(self, blueprint, command, _):
        table = self.wrap_table(blueprint)

        return "ALTER TABLE %s DROP CONSTRAINT %s" % (table, command.index)

    def compile_drop_index(self, blueprint, command, _):
        return "DROP INDEX %s" % command.index

    def compile_drop_foreign(self, blueprint, command, _):
        table = self.wrap_table(blueprint)

        return "ALTER TABLE %s DROP CONSTRAINT %s" % (table, command.index)

    def compile_rename(self, blueprint, command, _):
        from_ = self.wrap_table(blueprint)

        return "ALTER TABLE %s RENAME TO %s" % (from_, self.wrap_table(command.to))

    def _type_char(self, column):
        return "CHAR(%s)" % column.length

    def _type_string(self, column):
        return "VARCHAR(%s)" % column.length

    def _type_text(self, column):
        return "TEXT"

    def _type_medium_text(self, column):
        return "TEXT"

    def _type_long_text(self, column):
        return "TEXT"

    def _type_integer(self, column):
        return "SERIAL" if column.auto_increment else "INTEGER"

    def _type_big_integer(self, column):
        return "BIGSERIAL" if column.auto_increment else "BIGINT"

    def _type_medium_integer(self, column):
        return "SERIAL" if column.auto_increment else "INTEGER"

    def _type_tiny_integer(self, column):
        return "SMALLSERIAL" if column.auto_increment else "SMALLINT"

    def _type_small_integer(self, column):
        return "SMALLSERIAL" if column.auto_increment else "SMALLINT"

    def _type_float(self, column):
        return self._type_double(column)

    def _type_double(self, column):
        return "DOUBLE PRECISION"

    def _type_decimal(self, column):
        return "DECIMAL(%s, %s)" % (column.total, column.places)

    def _type_boolean(self, column):
        return "BOOLEAN"

    def _type_enum(self, column):
        allowed = list(map(lambda a: "'%s'" % a, column.allowed))

        return 'VARCHAR(255) CHECK ("%s" IN (%s))' % (column.name, ", ".join(allowed))

    def _type_json(self, column):
        return "JSON"

    def _type_date(self, column):
        return "DATE"

    def _type_datetime(self, column):
        return "TIMESTAMP(6) WITHOUT TIME ZONE"

    def _type_time(self, column):
        return "TIME(6) WITHOUT TIME ZONE"

    def _type_timestamp(self, column):
        if column.use_current:
            return "TIMESTAMP(6) WITHOUT TIME ZONE DEFAULT CURRENT_TIMESTAMP(6)"

        return "TIMESTAMP(6) WITHOUT TIME ZONE"

    def _type_binary(self, column):
        return "BYTEA"

    def _modify_nullable(self, blueprint, column):
        if column.get("nullable"):
            return " NULL"

        return " NOT NULL"

    def _modify_default(self, blueprint, column):
        if column.get("default") is not None:
            return " DEFAULT %s" % self._get_default_value(column.default)

        return ""

    def _modify_increment(self, blueprint, column):
        if column.type in self._serials and column.auto_increment:
            return " PRIMARY KEY"

        return ""

    def _get_dbal_column_type(self, type_):
        """
        Get the dbal column type.

        :param type_: The fluent type
        :type type_: str

        :rtype: str
        """
        type_ = type_.lower()

        if type_ == "enum":
            return "string"

        return super(PostgresSchemaGrammar, self)._get_dbal_column_type(type_)



# -*- coding: utf-8 -*-

from ..query.expression import QueryExpression


class Grammar(object):

    marker = "?"

    def __init__(self, marker=None):
        self._table_prefix = ""

        if marker:
            self.marker = marker

    def wrap_list(self, values):
        return list(map(self.wrap, values))

    def wrap_table(self, table):
        if self.is_expression(table):
            return self.get_value(table)

        return self.wrap(self._table_prefix + str(table), True)

    def wrap(self, value, prefix_alias=False):
        if self.is_expression(value):
            return self.get_value(value)

        # If the value being wrapped has a column alias we will need
        # to separate out the pieces so we can wrap each of the segments
        # of the expression on it own, and then joins them
        # both back together with the "as" connector.
        if value.lower().find(" as ") >= 0:
            segments = value.split(" ")

            if prefix_alias:
                segments[2] = self._table_prefix + segments[2]

            return "%s AS %s" % (self.wrap(segments[0]), self._wrap_value(segments[2]))

        wrapped = []

        segments = value.split(".")

        # If the value is not an aliased table expression, we'll just wrap it like
        # normal, so if there is more than one segment, we will wrap the first
        # segments as if it was a table and the rest as just regular values.
        for key, segment in enumerate(segments):
            if key == 0 and len(segments) > 1:
                wrapped.append(self.wrap_table(segment))
            else:
                wrapped.append(self._wrap_value(segment))

        return ".".join(wrapped)

    def _wrap_value(self, value):
        if value == "*":
            return value

        return '"%s"' % value.replace('"', '""')

    def columnize(self, columns):
        return ", ".join(map(self.wrap, columns))

    def parameterize(self, values):
        return ", ".join(map(self.parameter, values))

    def parameter(self, value):
        if self.is_expression(value):
            return self.get_value(value)

        return self.get_marker()

    def get_value(self, expression):
        return expression.get_value()

    def is_expression(self, value):
        return isinstance(value, QueryExpression)

    def get_date_format(self):
        return "%Y-%m-%d %H:%M:%S.%f"

    def get_table_prefix(self):
        return self._table_prefix

    def set_table_prefix(self, prefix):
        self._table_prefix = prefix

        return self

    def get_marker(self):
        return self.marker



from typing import TYPE_CHECKING
from lazy_imports import LazyImporter
import sys

_import_structure = {
    "src": [
        "FormatDuration",
        "Sleep",
        "Strftime",
        "Strptime",
        "Now",
        "DailyTimeBetween"
    ],
}

if TYPE_CHECKING:
    from .src import (
        FormatDuration,
        Sleep,
        Strftime,
        Strptime,
        Now,
        DailyTimeBetween
    )
else:
    sys.modules[__name__] = LazyImporter(
        __name__,
        globals()["__file__"],
        _import_structure,
        extra_objects={},
    )





# -*- coding: utf-8 -*-

import os
import glob
import inflection
import logging
from pygments import highlight
from pygments.lexers.sql import SqlLexer
from ..utils import decode, load_module
from ..utils.command_formatter import CommandFormatter


class MigratorHandler(logging.NullHandler):
    def __init__(self, level=logging.DEBUG):
        super(MigratorHandler, self).__init__(level)

        self.queries = []

    def handle(self, record):
        self.queries.append(record.query)


class Migrator(object):
    def __init__(self, repository, resolver):
        """
        :type repository: DatabaseMigrationRepository
        :type resolver: orator.database_manager.DatabaseManager
        """
        self._repository = repository
        self._resolver = resolver
        self._connection = None
        self._notes = []

    def run(self, path, pretend=False):
        """
        Run the outstanding migrations for a given path.

        :param path: The path
        :type path: str
        :param pretend: Whether we execute the migrations as dry-run
        :type pretend: bool
        """
        self._notes = []

        files = self._get_migration_files(path)

        ran = self._repository.get_ran()

        migrations = [f for f in files if f not in ran]

        self.run_migration_list(path, migrations, pretend)

    def run_migration_list(self, path, migrations, pretend=False):
        """
        Run a list of migrations.

        :type migrations: list

        :type pretend: bool
        """
        if not migrations:
            self._note("<info>Nothing to migrate</info>")

            return

        batch = self._repository.get_next_batch_number()

        for f in migrations:
            self._run_up(path, f, batch, pretend)

    def _run_up(self, path, migration_file, batch, pretend=False):
        """
        Run "up" a migration instance.

        :type migration_file: str

        :type batch: int

        :type pretend: bool
        """
        migration = self._resolve(path, migration_file)

        if pretend:
            return self._pretend_to_run(migration, "up")

        if migration.transactional:
            with migration.db.transaction():
                migration.up()
        else:
            migration.up()

        self._repository.log(migration_file, batch)

        self._note(
            decode("[<info>OK</>] <info>Migrated</info> ")
            + "<fg=cyan>%s</>" % migration_file
        )

    def rollback(self, path, pretend=False):
        """
        Rollback the last migration operation.

        :param path: The path
        :type path: str

        :param pretend: Whether we execute the migrations as dry-run
        :type pretend: bool

        :rtype: int
        """
        self._notes = []

        migrations = self._repository.get_last()

        if not migrations:
            self._note("<info>Nothing to rollback.</info>")

            return len(migrations)

        for migration in migrations:
            self._run_down(path, migration, pretend)

        return len(migrations)

    def reset(self, path, pretend=False):
        """
        Rolls all of the currently applied migrations back.

        :param path: The path
        :type path: str

        :param pretend: Whether we execute the migrations as dry-run
        :type pretend: bool

        :rtype: count
        """
        self._notes = []

        migrations = sorted(self._repository.get_ran(), reverse=True)

        count = len(migrations)

        if count == 0:
            self._note("<info>Nothing to rollback.</info>")
        else:
            for migration in migrations:
                self._run_down(path, {"migration": migration}, pretend)

        return count

    def _run_down(self, path, migration, pretend=False):
        """
        Run "down" a migration instance.
        """
        migration_file = migration["migration"]

        instance = self._resolve(path, migration_file)

        if pretend:
            return self._pretend_to_run(instance, "down")

        if instance.transactional:
            with instance.db.transaction():
                instance.down()
        else:
            instance.down()

        self._repository.delete(migration)

        self._note(
            decode("[<info>OK</>] <info>Rolled back</info> ")
            + "<fg=cyan>%s</>" % migration_file
        )

    def _get_migration_files(self, path):
        """
        Get all of the migration files in a given path.

        :type path: str

        :rtype: list
        """
        files = glob.glob(os.path.join(path, "[0-9]*_*.py"))

        if not files:
            return []

        files = list(map(lambda f: os.path.basename(f).replace(".py", ""), files))

        files = sorted(files)

        return files

    def _pretend_to_run(self, migration, method):
        """
        Pretend to run the migration.

        :param migration: The migration
        :type migration: orator.migrations.migration.Migration

        :param method: The method to execute
        :type method: str
        """
        self._note("")
        names = []
        for query in self._get_queries(migration, method):
            name = migration.__class__.__name__
            bindings = None

            if isinstance(query, tuple):
                query, bindings = query

            query = highlight(query, SqlLexer(), CommandFormatter()).strip()

            if bindings:
                query = (query, bindings)

            if name not in names:
                self._note("[<info>{}</info>]".format(name))
                names.append(name)

            self._note(query)

    def _get_queries(self, migration, method):
        """
        Get all of the queries that would be run for a migration.

        :param migration: The migration
        :type migration: orator.migrations.migration.Migration

        :param method: The method to execute
        :type method: str

        :rtype: list
        """
        connection = migration.get_connection()
        db = connection

        with db.pretend():
            getattr(migration, method)()

        return db.get_logged_queries()

    def _resolve(self, path, migration_file):
        """
        Resolve a migration instance from a file.

        :param migration_file: The migration file
        :type migration_file: str

        :rtype: orator.migrations.migration.Migration
        """
        name = "_".join(migration_file.split("_")[4:])
        migration_file = os.path.join(path, "%s.py" % migration_file)

        # Loading parent module
        parent = os.path.join(path, "__init__.py")
        if not os.path.exists(parent):
            with open(parent, "w"):
                pass

        load_module("migrations", parent)

        # Loading module
        mod = load_module("migrations.%s" % name, migration_file)

        klass = getattr(mod, inflection.camelize(name))

        instance = klass()
        instance.set_connection(self.get_repository().get_connection())

        return instance

    def _note(self, message):
        """
        Add a note to the migrator.

        :param message: The message
        :type message: str
        """
        self._notes.append(message)

    def resolve_connection(self, connection):
        return self._resolver.connection(connection)

    def set_connection(self, name):
        if name is not None:
            self._resolver.set_default_connection(name)

        self._repository.set_source(name)

        self._connection = name

    def get_repository(self):
        return self._repository

    def repository_exists(self):
        return self._repository.repository_exists()

    def get_notes(self):
        return self._notes



# -*- coding: utf-8 -*-

from .connection import Connection
from .mysql_connection import MySQLConnection
from .postgres_connection import PostgresConnection
from .sqlite_connection import SQLiteConnection



# -*- coding: utf-8 -*-

from collections import OrderedDict
from .platform import Platform
from .keywords.sqlite_keywords import SQLiteKeywords
from ..table import Table
from ..index import Index
from ..column import Column
from ..identifier import Identifier
from ..foreign_key_constraint import ForeignKeyConstraint
from ..exceptions import DBALException


class SQLitePlatform(Platform):

    INTERNAL_TYPE_MAPPING = {
        "boolean": "boolean",
        "tinyint": "boolean",
        "smallint": "smallint",
        "mediumint": "integer",
        "int": "integer",
        "integer": "integer",
        "serial": "integer",
        "bigint": "bigint",
        "bigserial": "bigint",
        "clob": "text",
        "tinytext": "text",
        "mediumtext": "text",
        "longtext": "text",
        "text": "text",
        "varchar": "string",
        "longvarchar": "string",
        "varchar2": "string",
        "nvarchar": "string",
        "image": "string",
        "ntext": "string",
        "char": "string",
        "date": "date",
        "datetime": "datetime",
        "timestamp": "datetime",
        "time": "time",
        "float": "float",
        "double": "float",
        "double precision": "float",
        "real": "float",
        "decimal": "decimal",
        "numeric": "decimal",
        "blob": "blob",
    }

    def get_list_table_columns_sql(self, table):
        table = table.replace(".", "__")

        return "PRAGMA table_info('%s')" % table

    def get_list_table_indexes_sql(self, table):
        table = table.replace(".", "__")

        return "PRAGMA index_list('%s')" % table

    def get_list_table_foreign_keys_sql(self, table):
        table = table.replace(".", "__")

        return "PRAGMA foreign_key_list('%s')" % table

    def get_pre_alter_table_index_foreign_key_sql(self, diff):
        """
        :param diff: The table diff
        :type diff: orator.dbal.table_diff.TableDiff

        :rtype: list
        """
        if not isinstance(diff.from_table, Table):
            raise DBALException(
                "Sqlite platform requires for alter table the table"
                "diff with reference to original table schema"
            )

        sql = []
        for index in diff.from_table.get_indexes().values():
            if not index.is_primary():
                sql.append(self.get_drop_index_sql(index, diff.name))

        return sql

    def get_post_alter_table_index_foreign_key_sql(self, diff):
        """
        :param diff: The table diff
        :type diff: orator.dbal.table_diff.TableDiff

        :rtype: list
        """
        if not isinstance(diff.from_table, Table):
            raise DBALException(
                "Sqlite platform requires for alter table the table"
                "diff with reference to original table schema"
            )

        sql = []

        if diff.new_name:
            table_name = diff.get_new_name()
        else:
            table_name = diff.get_name(self)

        for index in self._get_indexes_in_altered_table(diff).values():
            if index.is_primary():
                continue

            sql.append(
                self.get_create_index_sql(index, table_name.get_quoted_name(self))
            )

        return sql

    def get_create_table_sql(self, table, create_flags=None):
        if not create_flags:
            create_flags = self.CREATE_INDEXES | self.CREATE_FOREIGNKEYS

        return super(SQLitePlatform, self).get_create_table_sql(table, create_flags)

    def _get_create_table_sql(self, table_name, columns, options=None):
        table_name = table_name.replace(".", "__")
        query_fields = self.get_column_declaration_list_sql(columns)

        if options.get("unique_constraints"):
            for name, definition in options["unique_constraints"].items():
                query_fields += ", %s" % self.get_unique_constraint_declaration_sql(
                    name, definition
                )

        if options.get("primary"):
            key_columns = options["primary"]
            query_fields += ", PRIMARY KEY(%s)" % ", ".join(key_columns)

        if options.get("foreign_keys"):
            for foreign_key in options["foreign_keys"]:
                query_fields += ", %s" % self.get_foreign_key_declaration_sql(
                    foreign_key
                )

        query = ["CREATE TABLE %s (%s)" % (table_name, query_fields)]

        if options.get("alter"):
            return query

        if options.get("indexes"):
            for index_def in options["indexes"].values():
                query.append(self.get_create_index_sql(index_def, table_name))

        if options.get("unique"):
            for index_def in options["unique"].values():
                query.append(self.get_create_index_sql(index_def, table_name))

        return query

    def get_foreign_key_declaration_sql(self, foreign_key):
        return super(SQLitePlatform, self).get_foreign_key_declaration_sql(
            ForeignKeyConstraint(
                foreign_key.get_quoted_local_columns(self),
                foreign_key.get_quoted_foreign_table_name(self).replace(".", "__"),
                foreign_key.get_quoted_foreign_columns(self),
                foreign_key.get_name(),
                foreign_key.get_options(),
            )
        )

    def get_advanced_foreign_key_options_sql(self, foreign_key):
        query = super(SQLitePlatform, self).get_advanced_foreign_key_options_sql(
            foreign_key
        )

        deferrable = (
            foreign_key.has_option("deferrable")
            and foreign_key.get_option("deferrable") is not False
        )
        if deferrable:
            query += " DEFERRABLE"
        else:
            query += " NOT DEFERRABLE"

        query += " INITIALLY"

        deferred = (
            foreign_key.has_option("deferred")
            and foreign_key.get_option("deferred") is not False
        )
        if deferred:
            query += " DEFERRED"
        else:
            query += " IMMEDIATE"

        return query

    def get_alter_table_sql(self, diff):
        """
        Get the ALTER TABLE SQL statement

        :param diff: The table diff
        :type diff: orator.dbal.table_diff.TableDiff

        :rtype: list
        """
        sql = self._get_simple_alter_table_sql(diff)
        if sql is not False:
            return sql

        from_table = diff.from_table
        if not isinstance(from_table, Table):
            raise DBALException(
                "SQLite platform requires for the alter table the table diff "
                "referencing the original table"
            )

        table = from_table.clone()
        columns = OrderedDict()
        old_column_names = OrderedDict()
        new_column_names = OrderedDict()
        column_sql = []
        for column_name, column in table.get_columns().items():
            column_name = column_name.lower()
            columns[column_name] = column
            old_column_names[column_name] = column.get_quoted_name(self)
            new_column_names[column_name] = column.get_quoted_name(self)

        for column_name, column in diff.removed_columns.items():
            column_name = column_name.lower()
            if column_name in columns:
                del columns[column_name]
                del old_column_names[column_name]
                del new_column_names[column_name]

        for old_column_name, column in diff.renamed_columns.items():
            old_column_name = old_column_name.lower()
            if old_column_name in columns:
                del columns[old_column_name]

            columns[column.get_name().lower()] = column

            if old_column_name in new_column_names:
                new_column_names[old_column_name] = column.get_quoted_name(self)

        for old_column_name, column_diff in diff.changed_columns.items():
            if old_column_name in columns:
                del columns[old_column_name]

            columns[column_diff.column.get_name().lower()] = column_diff.column

            if old_column_name in new_column_names:
                new_column_names[old_column_name] = column_diff.column.get_quoted_name(
                    self
                )

        for column_name, column in diff.added_columns.items():
            columns[column_name.lower()] = column

        table_sql = []

        data_table = Table("__temp__" + table.get_name())
        new_table = Table(
            table.get_quoted_name(self),
            columns,
            self._get_primary_index_in_altered_table(diff),
            self._get_foreign_keys_in_altered_table(diff),
            table.get_options(),
        )
        new_table.add_option("alter", True)

        sql = self.get_pre_alter_table_index_foreign_key_sql(diff)
        sql.append(
            "CREATE TEMPORARY TABLE %s AS SELECT %s FROM %s"
            % (
                data_table.get_quoted_name(self),
                ", ".join(old_column_names.values()),
                table.get_quoted_name(self),
            )
        )
        sql.append(self.get_drop_table_sql(from_table))

        sql += self.get_create_table_sql(new_table)
        sql.append(
            "INSERT INTO %s (%s) SELECT %s FROM %s"
            % (
                new_table.get_quoted_name(self),
                ", ".join(new_column_names.values()),
                ", ".join(old_column_names.values()),
                data_table.get_name(),
            )
        )
        sql.append(self.get_drop_table_sql(data_table))

        sql += self.get_post_alter_table_index_foreign_key_sql(diff)

        return sql

    def _get_simple_alter_table_sql(self, diff):
        for old_column_name, column_diff in diff.changed_columns.items():
            if (
                not isinstance(column_diff.from_column, Column)
                or not isinstance(column_diff.column, Column)
                or not column_diff.column.get_autoincrement()
                or column_diff.column.get_type().lower() != "integer"
            ):
                continue

            if not column_diff.has_changed("type") and not column_diff.has_changed(
                "unsigned"
            ):
                del diff.changed_columns[old_column_name]

                continue

            from_column_type = column_diff.column.get_type()

            if from_column_type == "smallint" or from_column_type == "bigint":
                del diff.changed_columns[old_column_name]

        if any(
            [
                not diff.renamed_columns,
                not diff.added_foreign_keys,
                not diff.added_indexes,
                not diff.changed_columns,
                not diff.changed_foreign_keys,
                not diff.changed_indexes,
                not diff.removed_columns,
                not diff.removed_foreign_keys,
                not diff.removed_indexes,
                not diff.renamed_indexes,
            ]
        ):
            return False

        table = Table(diff.name)

        sql = []
        table_sql = []
        column_sql = []

        for column in diff.added_columns.values():
            field = {"unique": None, "autoincrement": None, "default": None}
            field.update(column.to_dict())

            type_ = field["type"]
            if (
                "column_definition" in field
                or field["autoincrement"]
                or field["unique"]
            ):
                return False
            elif (
                type_ == "datetime"
                and field["default"] == self.get_current_timestamp_sql()
            ):
                return False
            elif type_ == "date" and field["default"] == self.get_current_date_sql():
                return False
            elif type_ == "time" and field["default"] == self.get_current_time_sql():
                return False

            field["name"] = column.get_quoted_name(self)
            if field["type"].lower() == "string" and field["length"] is None:
                field["length"] = 255

            sql.append(
                "ALTER TABLE "
                + table.get_quoted_name(self)
                + " ADD COLUMN "
                + self.get_column_declaration_sql(field["name"], field)
            )

        if diff.new_name is not False:
            new_table = Identifier(diff.new_name)
            sql.append(
                "ALTER TABLE "
                + table.get_quoted_name(self)
                + " RENAME TO "
                + new_table.get_quoted_name(self)
            )

        return sql

    def _get_indexes_in_altered_table(self, diff):
        """
        :param diff: The table diff
        :type diff: orator.dbal.table_diff.TableDiff

        :rtype: dict
        """
        indexes = diff.from_table.get_indexes()
        column_names = self._get_column_names_in_altered_table(diff)

        for key, index in OrderedDict([(k, v) for k, v in indexes.items()]).items():
            for old_index_name, renamed_index in diff.renamed_indexes.items():
                if key.lower() == old_index_name.lower():
                    del indexes[key]

            changed = False
            index_columns = []
            for column_name in index.get_columns():
                normalized_column_name = column_name.lower()
                if normalized_column_name not in column_names:
                    del indexes[key]
                    break
                else:
                    index_columns.append(column_names[normalized_column_name])
                    if column_name != column_names[normalized_column_name]:
                        changed = True

            if changed:
                indexes[key] = Index(
                    index.get_name(),
                    index_columns,
                    index.is_unique(),
                    index.is_primary(),
                    index.get_flags(),
                )

            for index in diff.removed_indexes.values():
                index_name = index.get_name().lower()
                if index_name and index_name in indexes:
                    del indexes[index_name]

            changed_indexes = (
                list(diff.changed_indexes.values())
                + list(diff.added_indexes.values())
                + list(diff.renamed_indexes.values())
            )
            for index in changed_indexes:
                index_name = index.get_name().lower()
                if index_name:
                    indexes[index_name] = index
                else:
                    indexes[len(indexes)] = index

        return indexes

    def _get_column_names_in_altered_table(self, diff):
        """
        :param diff: The table diff
        :type diff: orator.dbal.table_diff.TableDiff

        :rtype: dict
        """
        columns = OrderedDict()

        for column_name, column in diff.from_table.get_columns().items():
            columns[column_name.lower()] = column.get_name()

        for column_name, column in diff.removed_columns.items():
            column_name = column_name.lower()
            if column_name in columns:
                del columns[column_name]

        for old_column_name, column in diff.renamed_columns.items():
            column_name = column.get_name()
            columns[old_column_name.lower()] = column_name
            columns[column_name.lower()] = column_name

        for old_column_name, column_diff in diff.changed_columns.items():
            column_name = column_diff.column.get_name()
            columns[old_column_name.lower()] = column_name
            columns[column_name.lower()] = column_name

        for column_name, column in diff.added_columns.items():
            columns[column_name.lower()] = column_name

        return columns

    def _get_foreign_keys_in_altered_table(self, diff):
        """
        :param diff: The table diff
        :type diff: orator.dbal.table_diff.TableDiff

        :rtype: dict
        """
        foreign_keys = diff.from_table.get_foreign_keys()
        column_names = self._get_column_names_in_altered_table(diff)

        for key, constraint in foreign_keys.items():
            changed = False
            local_columns = []
            for column_name in constraint.get_local_columns():
                normalized_column_name = column_name.lower()
                if normalized_column_name not in column_names:
                    del foreign_keys[key]
                    break
                else:
                    local_columns.append(column_names[normalized_column_name])
                    if column_name != column_names[normalized_column_name]:
                        changed = True

            if changed:
                foreign_keys[key] = ForeignKeyConstraint(
                    local_columns,
                    constraint.get_foreign_table_name(),
                    constraint.get_foreign_columns(),
                    constraint.get_name(),
                    constraint.get_options(),
                )

        for constraint in diff.removed_foreign_keys:
            constraint_name = constraint.get_name().lower()
            if constraint_name and constraint_name in foreign_keys:
                del foreign_keys[constraint_name]

        foreign_keys_diff = diff.changed_foreign_keys + diff.added_foreign_keys
        for constraint in foreign_keys_diff:
            constraint_name = constraint.get_name().lower()
            if constraint_name:
                foreign_keys[constraint_name] = constraint
            else:
                foreign_keys[len(foreign_keys)] = constraint

        return foreign_keys

    def _get_primary_index_in_altered_table(self, diff):
        """
        :param diff: The table diff
        :type diff: orator.dbal.table_diff.TableDiff

        :rtype: dict
        """
        primary_index = {}

        for index in self._get_indexes_in_altered_table(diff).values():
            if index.is_primary():
                primary_index = {index.get_name(): index}

        return primary_index

    def supports_foreign_key_constraints(self):
        return True

    def get_boolean_type_declaration_sql(self, column):
        return "BOOLEAN"

    def get_integer_type_declaration_sql(self, column):
        return "INTEGER" + self._get_common_integer_type_declaration_sql(column)

    def get_bigint_type_declaration_sql(self, column):
        # SQLite autoincrement is implicit for INTEGER PKs, but not for BIGINT fields.
        if not column.get("autoincrement", False):
            return self.get_integer_type_declaration_sql(column)

        return "BIGINT" + self._get_common_integer_type_declaration_sql(column)

    def get_tinyint_type_declaration_sql(self, column):
        # SQLite autoincrement is implicit for INTEGER PKs, but not for TINYINT fields.
        if not column.get("autoincrement", False):
            return self.get_integer_type_declaration_sql(column)

        return "TINYINT" + self._get_common_integer_type_declaration_sql(column)

    def get_smallint_type_declaration_sql(self, column):
        # SQLite autoincrement is implicit for INTEGER PKs, but not for SMALLINT fields.
        if not column.get("autoincrement", False):
            return self.get_integer_type_declaration_sql(column)

        return "SMALLINT" + self._get_common_integer_type_declaration_sql(column)

    def get_mediumint_type_declaration_sql(self, column):
        # SQLite autoincrement is implicit for INTEGER PKs, but not for MEDIUMINT fields.
        if not column.get("autoincrement", False):
            return self.get_integer_type_declaration_sql(column)

        return "MEDIUMINT" + self._get_common_integer_type_declaration_sql(column)

    def get_datetime_type_declaration_sql(self, column):
        return "DATETIME"

    def get_date_type_declaration_sql(self, column):
        return "DATE"

    def get_time_type_declaration_sql(self, column):
        return "TIME"

    def _get_common_integer_type_declaration_sql(self, column):
        # sqlite autoincrement is implicit for integer PKs, but not when the field is unsigned
        if not column.get("autoincrement", False):
            return ""

        if not column.get("unsigned", False):
            return " UNSIGNED"

        return ""

    def get_varchar_type_declaration_sql_snippet(self, length, fixed):
        if fixed:
            return "CHAR(%s)" % length if length else "CHAR(255)"
        else:
            return "VARCHAR(%s)" % length if length else "TEXT"

    def get_blob_type_declaration_sql(self, column):
        return "BLOB"

    def get_clob_type_declaration_sql(self, column):
        return "CLOB"

    def get_column_options(self):
        return ["pk"]

    def _get_reserved_keywords_class(self):
        return SQLiteKeywords



from bagbag import * 

class FearAndGreedIndexResult():
    def __init__(self) -> None:
        self.Score:int = None 
        self.Status:str = None 
        self.NextUpdateRemainSeconds:int = None 
    
    def __repr__(self) -> str:
        # ipdb.set_trace()
        # Lg.Trace(self.__dir__())
        # Lg.Trace([i for i in filter(lambda attr: attr[0].isupper() and type(getattr(self, attr)) in [int, float, str], self.__dir__())])
        return f"{self.__class__.__name__}(%s)" % ' '.join([f"{attr}={getattr(self, attr)}" for attr in filter(lambda attr: attr[0].isupper() and type(getattr(self, attr)) in [int, float, str], self.__dir__())])

    def __str__(self) -> str:
        return self.__repr__()

def FearAndGreedIndex() -> FearAndGreedIndexResult:
    html = Http.Get("https://alternative.me/crypto/fear-and-greed-index/").Content
    x = Tools.XPath(html)

    status = str(x.Find("/html/body/div/main/section/div/div[3]/div[2]/div/div/div[1]/div[1]/div[2]").Text())
    score = int(x.Find("/html/body/div/main/section/div/div[3]/div[2]/div/div/div[1]/div[2]/div").Text())
    nextupdate = int(x.Find("//countdown").Attribute(":time").split(" ")[0])

    # Lg.Trace(type(value))

    fagir = FearAndGreedIndexResult()
    fagir.Score = score
    fagir.Status = status
    fagir.NextUpdateRemainSeconds = nextupdate

    return fagir

if __name__ == "__main__":
    f = FearAndGreedIndex()
    Lg.Trace(f)


from .... import Http, Json, Time, Tools, String

#print("load " + '/'.join(__file__.split('/')[-2:]))

# https://github.com/binance/binance-spot-api-docs/blob/master/rest-api.md

class CoinsPairPrice():
    def __init__(self, pair:str, price:float, time:float) -> None:
        self.Pair:str = pair 
        self.Price:float = price  
        self.Time:float = time
    
    def __repr__(self) -> str:
        return f"CoinsPairPrice(Pair={self.Pair} Price={self.Price} Time={self.Time})"

    def __str__(self) -> str:
        return self.__repr__()

servertime = None
rl = Tools.RateLimit("1/s")

def ServerTime() -> float:
    global servertime 
    if servertime == None:
        servertime = Json.Loads(Http.Get("https://api.binance.com/api/v3/time").Content)['serverTime'] / 1000
    timegap = Time.Now() - servertime

    return Time.Now() - timegap

def GetPrice(pair:str|list="BTCUSDT") -> CoinsPairPrice | list[CoinsPairPrice]:
    """
    The function `GetPrice` retrieves the current price of a specified cryptocurrency pair or a list of
    pairs from the Binance API.
    
    :param pair: For example: BTCUSDT. The trading pair(s) for which you want to retrieve the current price(s). 
    It can be a string for a single pair or a list of strings for multiple pairs. If no pair is specified, the
    function will return the prices for all available trading pairs
    :type pair: str|list
    :return: The function `GetPrice` returns either a `CoinsPairPrice` object or a list of
    `CoinsPairPrice` objects depending on the input parameter `pair`. If `pair` is a string, a single
    `CoinsPairPrice` object is returned. If `pair` is a list, a list of `CoinsPairPrice` objects is
    returned. If `pair` is `None`,
    """
    rl.Take()
    
    url = "https://api.binance.com/api/v3/ticker/price"
    if pair != None:
        if type(pair) == str:
            url = url + "?symbol=" + pair
        elif type(pair) == list:
            url = url + "?symbols=" + String(Json.Dumps(pair).replace(" ", "")).URLEncode()  
        else:
            raise Exception("不合适的pair类型")
    
    resp = Http.Get(url)
    if resp.StatusCode != 200:
        raise Exception(f"状态码不为200: {resp.StatusCode}. 详见: https://github.com/binance/binance-spot-api-docs/blob/master/rest-api.md")
    
    st = ServerTime()

    content = Json.Loads(resp.Content)

    if pair != None and type(pair) == str:
        return CoinsPairPrice(content['symbol'], content['price'], st)

    resp = []
    for p in content:
        resp.append(CoinsPairPrice(p['symbol'], p['price'], st))

    return resp



# -*- coding: utf-8 -*-

from .keyword_list import KeywordList


class SQLiteKeywords(KeywordList):

    KEYWORDS = [
        "ABORT",
        "ACTION",
        "ADD",
        "AFTER",
        "ALL",
        "ALTER",
        "ANALYZE",
        "AND",
        "AS",
        "ASC",
        "ATTACH",
        "AUTOINCREMENT",
        "BEFORE",
        "BEGIN",
        "BETWEEN",
        "BY",
        "CASCADE",
        "CASE",
        "CAST",
        "CHECK",
        "COLLATE",
        "COLUMN",
        "COMMIT",
        "CONFLICT",
        "CONSTRAINT",
        "CREATE",
        "CROSS",
        "CURRENT_DATE",
        "CURRENT_TIME",
        "CURRENT_TIMESTAMP",
        "DATABASE",
        "DEFAULT",
        "DEFERRABLE",
        "DEFERRED",
        "DELETE",
        "DESC",
        "DETACH",
        "DISTINCT",
        "DROP",
        "EACH",
        "ELSE",
        "END",
        "ESCAPE",
        "EXCEPT",
        "EXCLUSIVE",
        "EXISTS",
        "EXPLAIN",
        "FAIL",
        "FOR",
        "FOREIGN",
        "FROM",
        "FULL",
        "GLOB",
        "GROUP",
        "HAVING",
        "IF",
        "IGNORE",
        "IMMEDIATE",
        "IN",
        "INDEX",
        "INDEXED",
        "INITIALLY",
        "INNER",
        "INSERT",
        "INSTEAD",
        "INTERSECT",
        "INTO",
        "IS",
        "ISNULL",
        "JOIN",
        "KEY",
        "LEFT",
        "LIKE",
        "LIMIT",
        "MATCH",
        "NATURAL",
        "NO",
        "NOT",
        "NOTNULL",
        "NULL",
        "OF",
        "OFFSET",
        "ON",
        "OR",
        "ORDER",
        "OUTER",
        "PLAN",
        "PRAGMA",
        "PRIMARY",
        "QUERY",
        "RAISE",
        "REFERENCES",
        "REGEXP",
        "REINDEX",
        "RELEASE",
        "RENAME",
        "REPLACE",
        "RESTRICT",
        "RIGHT",
        "ROLLBACK",
        "ROW",
        "SAVEPOINT",
        "SELECT",
        "SET",
        "TABLE",
        "TEMP",
        "TEMPORARY",
        "THEN",
        "TO",
        "TRANSACTION",
        "TRIGGER",
        "UNION",
        "UNIQUE",
        "UPDATE",
        "USING",
        "VACUUM",
        "VALUES",
        "VIEW",
        "VIRTUAL",
        "WHEN",
        "WHERE",
    ]

    def get_name(self):
        return "SQLite"



# -*- coding: utf-8 -*-

from wrapt import ObjectProxy


class Result(ObjectProxy):

    _results = None
    _relation = None
    _parent = None
    _kwargs = None

    def __init__(self, result, relation, parent, **kwargs):
        """
        :param query: A Builder instance
        :type query: orm.orator.Builder

        :param parent: The parent model
        :type parent: Model
        """
        super(Result, self).__init__(result)

        self._results = result
        self._relation = relation
        self._parent = parent
        self._kwargs = kwargs

    def __call__(self, *args, **kwargs):
        return self._relation.new_instance(self._parent, **self._kwargs)



from typing import TYPE_CHECKING
from lazy_imports import LazyImporter
import sys

_import_structure = {
    "OfficialAccountVertify": [
        "OfficialAccountVertify"
    ],
    "CoinsPrice_src": [
        "GetPrice"
    ]
}

if TYPE_CHECKING:
    from .CoinsPrice_src import GetPrice
    from .OfficialAccountVertify import OfficialAccountVertify
else:
    sys.modules[__name__] = LazyImporter(
        __name__,
        globals()["__file__"],
        _import_structure,
        extra_objects={},
    )





import json
import io
import os
import html_to_json
import demjson3

#print("load json")

def Dumps(obj, indent=4, ensure_ascii=False) -> str:
    """
    It takes a Python object and returns a JSON string
    
    :param obj: The object to be serialized
    :param indent: This is the number of spaces to indent for each level. If it is None, that
    will insert newlines but won't indent the new lines, defaults to 4 (optional)
    :param ensure_ascii: If True, all non-ASCII characters in the output are escaped with \\uXXXX
    sequences, and the result is a str instance consisting of ASCII characters only. If False, some
    chunks written to fp may be unicode instances. This usually happens because the input contains
    unicode strings or the, defaults to False (optional)
    :return: A string
    """
    return json.dumps(obj, indent=indent, ensure_ascii=ensure_ascii)

def Loads(s:str|io.TextIOWrapper) -> list | dict:
    """
    The function Loads can load JSON or HTML data from 
    1. a file (file path)
    2. a string
    3. a file object
    
    :param s: The parameter `s` can be either a string or a file object (`io.TextIOWrapper`). It
    represents the JSON data or HTML data that needs to be loaded
    :type s: str|io.TextIOWrapper
    :return: The function `Loads` returns a list or a dictionary depending on the input provided. If the
    input is a file object of type `io.TextIOWrapper`, it reads the contents of the file and returns a
    dictionary or a list after parsing the JSON data. If the input is a string, it checks if the string
    starts with `[` or `{` and returns a dictionary or a list after
    """
    if type(s) == io.TextIOWrapper:
        return demjson3.decode(s.read())
    
    if type(s) == str:
        if s.startswith('[') or s.startswith('{'):
            return demjson3.decode(s)
        elif os.path.exists(s):
            data = open(s).read()
            try:
                return demjson3.decode(data)
            except:
                return html_to_json.convert(data)
        else:
            return html_to_json.convert(s)

def Valid(json_string:str) -> bool:
    """
    检查一个 JSON 字符串是否合法

    参数:
        json_string (str): 要检查的 JSON 字符串

    返回:
        bool: 如果 JSON 字符串合法则返回 True，否则返回 False
    """
    try:
        json.loads(json_string)
        return True
    except ValueError:
        return False

def ExtraValueByKey(obj:list|dict, key:str) -> list:
    """Recursively fetch values from nested JSON."""
    arr = []

    def extract(obj, arr, key):
        """Recursively search for values of key in JSON tree."""
        if isinstance(obj, dict):
            for k, v in obj.items():
                if isinstance(v, (dict, list)):
                    extract(v, arr, key)

                if k == key:
                    arr.append(v)
                    
        elif isinstance(obj, list):
            for item in obj:
                extract(item, arr, key)
        return arr

    values = extract(obj, arr, key)
    return values

def DeleteKeyContainString(obj:list|dict, target_string:str) -> dict|list:
    """
    遍历字典并删除包含特定字符串的键值对

    参数:
    d (dict): 需要遍历的字典
    target_string (str): 需要查找的字符串

    返回:
    dict: 删除包含目标字符串的键值对后的字典
    """
    keys_to_delete = [key for key in obj if target_string in key]

    for key in keys_to_delete:
        del obj[key]

    for key, value in obj.items():
        if isinstance(value, dict):
            DeleteKeyContainString(value, target_string)

    return obj

def DeleteKeyMatchString(obj:list|dict, target_string:str) -> dict|list:
    """
    遍历字典并删除等于特定字符串的键值对

    参数:
    d (dict): 需要遍历的字典
    target_string (str): 需要查找的字符串

    返回:
    dict: 删除等于目标字符串的键值对后的字典
    """
    keys_to_delete = [key for key in obj if target_string == key]

    for key in keys_to_delete:
        del obj[key]

    for key, value in obj.items():
        if isinstance(value, dict):
            DeleteKeyContainString(value, target_string)

    return obj

if __name__ == "__main__":
    # j = Dumps({1: 3, 4: 5})
    # print(j)

    # d = Loads(j)
    # print(d)

    # print(type(d))

    # ------------

    # data = {
    #     "key": {
    #         "key": [
    #             {
    #                 "a": "b"
    #             },
    #             {
    #                 "key": "123"
    #             }
    #         ]
    #     }
    # }

    # print(ExtraValueByKey(data, "key"))

    html_string = """<head>
    <title>Test site</title>
    <meta charset="UTF-8"></head>"""

    print(Loads(html_string))